;; Constraint definitions for Altera Nios II.
;; Copyright (C) 2012 Free Software Foundation, Inc.
;; Contributed by Chung-Lin Tang <cltang@codesourcery.com>
;;
;; This file is part of GCC.
;;
;; GCC is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 3, or (at your option)
;; any later version.
;;
;; GCC is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with GCC; see the file COPYING3.  If not see
;; <http://www.gnu.org/licenses/>.

;; We use the following constraint letters for constants
;;
;;  I: -32768 to -32767
;;  J: 0 to 65535
;;  K: $nnnn0000 for some nnnn
;;  L: 0 to 31 (for shift counts)
;;  M: 0
;;  N: 0 to 255 (for custom instruction numbers)
;;  O: 0 to 31 (for control register numbers)
;;
;; We use the following built-in register classes:
;;
;;  r: general purpose register (r0..r31)
;;  m: memory operand
;;
;; Plus, we define the following constraint strings:
;;
;;  S: symbol that is in the "small data" area
;;  Dnn: Dnn_REG (just rnn)

;; Register constraints

(define_register_constraint "j" "SIB_REGS"
  "A register suitable for an indirect sibcall.")

;; These are documented for use in inline asm.
(define_register_constraint "D00" "D00_REG" "Hard register 0.")
(define_register_constraint "D01" "D01_REG" "Hard register 1.")
(define_register_constraint "D02" "D02_REG" "Hard register 2.")
(define_register_constraint "D03" "D03_REG" "Hard register 3.")
(define_register_constraint "D04" "D04_REG" "Hard register 4.")
(define_register_constraint "D05" "D05_REG" "Hard register 5.")
(define_register_constraint "D06" "D06_REG" "Hard register 6.")
(define_register_constraint "D07" "D07_REG" "Hard register 7.")
(define_register_constraint "D08" "D08_REG" "Hard register 8.")
(define_register_constraint "D09" "D09_REG" "Hard register 9.")
(define_register_constraint "D10" "D10_REG" "Hard register 10.")
(define_register_constraint "D11" "D11_REG" "Hard register 11.")
(define_register_constraint "D12" "D12_REG" "Hard register 12.")
(define_register_constraint "D13" "D13_REG" "Hard register 13.")
(define_register_constraint "D14" "D14_REG" "Hard register 14.")
(define_register_constraint "D15" "D15_REG" "Hard register 15.")
(define_register_constraint "D16" "D16_REG" "Hard register 16.")
(define_register_constraint "D17" "D17_REG" "Hard register 17.")
(define_register_constraint "D18" "D18_REG" "Hard register 18.")
(define_register_constraint "D19" "D19_REG" "Hard register 19.")
(define_register_constraint "D20" "D20_REG" "Hard register 20.")
(define_register_constraint "D21" "D21_REG" "Hard register 21.")
(define_register_constraint "D22" "D22_REG" "Hard register 22.")
(define_register_constraint "D23" "D23_REG" "Hard register 23.")
(define_register_constraint "D24" "D24_REG" "Hard register 24.")
(define_register_constraint "D25" "D25_REG" "Hard register 25.")
(define_register_constraint "D26" "D26_REG" "Hard register 26.")
(define_register_constraint "D27" "D27_REG" "Hard register 27.")
(define_register_constraint "D28" "D28_REG" "Hard register 28.")
(define_register_constraint "D29" "D29_REG" "Hard register 29.")
(define_register_constraint "D30" "D30_REG" "Hard register 30.")
(define_register_constraint "D31" "D31_REG" "Hard register 31.")

;; Integer constraints

(define_constraint "I"
  "A signed 16-bit constant (for arithmetic instructions)."
  (and (match_code "const_int")
       (match_test "SMALL_INT (ival)")))

(define_constraint "J"
  "An unsigned 16-bit constant (for logical instructions)."
  (and (match_code "const_int")
       (match_test "SMALL_INT_UNSIGNED (ival)")))

(define_constraint "K"
  "An unsigned 16-bit high constant (for logical instructions)."
  (and (match_code "const_int")
       (match_test "UPPER16_INT (ival)")))

(define_constraint "L"
  "An unsigned 5-bit constant (for shift counts)."
  (and (match_code "const_int")
       (match_test "ival >= 0 && ival <= 31")))

(define_constraint "M"
  "Integer zero."
  (and (match_code "const_int")
       (match_test "ival == 0")))

(define_constraint "N"
  "An unsigned 8-bit constant (for custom instruction codes)."
  (and (match_code "const_int")
       (match_test "ival >= 0 && ival <= 255")))

(define_constraint "O"
  "An unsigned 5-bit constant (for control register numbers)."
  (and (match_code "const_int")
       (match_test "ival >= 0 && ival <= 31")))

(define_constraint "S"
  "An immediate stored in small data, accessible by GP."
  (and (match_code "symbol_ref")
       (match_test "SYMBOL_REF_IN_NIOS2_SMALL_DATA_P (op)")))
